"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
from django.urls import resolve
from django.contrib.auth.models import User

from django_oci import settings
from django_oci.utils import get_server
from django_oci.models import Repository

from rest_framework.authtoken.models import Token
from rest_framework.response import Response

from django.middleware import cache

from datetime import datetime
import uuid
import base64
import re
import time
import jwt


def is_authenticated(
    request, repository=None, must_be_owner=True, repository_exists=True
):
    """
    Function to check if a request is authenticated, a repository and the request is required.
    Returns a boolean to indicate if the user is authenticated, and a response with
    the challenge if not. If allow_if_private is True, we only allow access to users
    that are owners or contributors, regardless of having a valid token.

    Arguments:
    ==========
    request (requests.Request)    : the Request object to inspect
    repository (str or Repository): the name of a repository or instance
    must_be_owner (bool)          : if must be owner is true, requires push
    reposity_exists (bool)        : flag to indicate that the repository exists.
    """
    # Derive the view name from the request PATH_INFO
    func, two, three = resolve(request.META["PATH_INFO"])
    view_name = "%s.%s" % (func.__module__, func.__name__)

    # If authentication is disabled, return the original view
    if settings.DISABLE_AUTHENTICATION or view_name not in settings.AUTHENTICATED_VIEWS:
        return True, None, None

    # Ensure repository is valid, only if provided
    name = repository
    if repository is not None and repository_exists and isinstance(repository, str):
        try:
            repository = Repository.objects.get(name=repository)
            name = repository.name
        except Repository.DoesNotExist:
            return False, Response(status=404), None

    # Case 2: Already has a jwt valid token
    is_valid, user = validate_jwt(request, repository, must_be_owner)
    if is_valid:
        return True, None, user

    # Case 3: False and response will return request for auth
    user = get_user(request)
    if not user:
        headers = {"Www-Authenticate": get_challenge(request, name)}
        return False, Response(status=401, headers=headers), user

    # Denied for any other reason
    return False, Response(status=403), user


def generate_jwt(username, scope, realm, repository):
    """Given a username, scope, realm, repository, and service, generate a jwt
    token to return to the user with a default expiration of 10 minutes.

    Arguments:
    ==========
    username (str)  : the user's username to add under "sub"
    scope (list)    : a list of scopes to require (e.g., ['push, pull'])
    realm (str)     : the authentication realm, typically <server>/auth
    repository (str): the repository name
    """
    # The jti expires after TOKEN_EXPIRES_SECONDS
    issued_at = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    filecache = cache.caches["django_oci_upload"]
    jti = str(uuid.uuid4())
    filecache.set(jti, "good", timeout=settings.TOKEN_EXPIRES_SECONDS)
    now = int(time.time())
    expires_at = now + settings.TOKEN_EXPIRES_SECONDS

    # import jwt and generate token
    # https://tools.ietf.org/html/rfc7519#section-4.1.5
    payload = {
        "iss": realm,  # auth endpoint
        "sub": username,
        "exp": expires_at,
        "nbf": now,
        "iat": now,
        "jti": jti,
        "access": [{"type": "repository", "name": repository, "actions": scope}],
    }
    token = jwt.encode(payload, settings.JWT_SERVER_SECRET, algorithm="HS256").decode(
        "utf-8"
    )
    return {
        "token": token,
        "expires_in": settings.TOKEN_EXPIRES_SECONDS,
        "issued_at": issued_at,
    }


def validate_jwt(request, repository, must_be_owner):
    """Given a jwt token, decode and validate

    Arguments:
    ==========
    request (requests.Request)    : the Request object to inspect
    repository (models.Repository): the repository instance
    must_be_owner (bool)          : if True, requires additional push scope
    """
    header = request.META.get("HTTP_AUTHORIZATION", "")
    if re.search("bearer", header, re.IGNORECASE):
        encoded = re.sub("bearer", "", header, flags=re.IGNORECASE).strip()

        # Any reason not valid will issue an error here
        try:
            decoded = jwt.decode(
                encoded, settings.JWT_SERVER_SECRET, algorithms=["HS256"]
            )
        except Exception as exc:
            print("jwt could no be decoded, %s" % exc)
            return False, None

        # Ensure that the jti is still valid
        filecache = cache.caches["django_oci_upload"]
        if not filecache.get(decoded.get("jti")) == "good":
            print("Filecache with jti not found.")
            return False, None

        # The user must exist
        try:
            user = User.objects.get(username=decoded.get("sub"))
        except User.DoesNotExist:
            print("Username %s not found" % decoded.get("sub"))
            return False, None

        # If a repository exists, the user must be an owner
        if (
            isinstance(repository, Repository)
            and (repository.private or must_be_owner)
            and user not in repository.owners.all()
            and user not in repository.contributors.all()
        ):
            print("Username %s not in repository owners" % decoded.get("sub"))
            return False, None

        # If repository is not defined and must be owner, no go
        if repository is None and must_be_owner:
            print("Repository is None and must be owner")
            return False, None

        # TODO: any validation needed for access type?
        requested_name = decoded.get("access", [{}])[0].get("name")
        if isinstance(repository, Repository) and repository.name != requested_name:
            print("Repository name is not equal to requested name.")
            return False, None

        # Do we have the correct permissions?
        requested_permission = decoded.get("access", [{}])[0].get("actions")
        if must_be_owner and "push" not in requested_permission:
            print("Must be owner and push not in requested permissions")
            return False, None
        return True, user

    return False, None


def get_user(request):
    """Given a request, read the Authorization header to get the base64 encoded
    username and token (password) which is a basic auth. If we return the user
    object, the user is successfully authenticated. Otherwise, return None.
    and the calling function should return Forbidden status.

    Arguments:
    ==========
    request (requests.Request)    : the Request object to inspect
    """
    header = request.META.get("HTTP_AUTHORIZATION", "")

    if re.search("basic", header, re.IGNORECASE):
        encoded = re.sub("basic", "", header, flags=re.IGNORECASE).strip()
        decoded = base64.b64decode(encoded).decode("utf-8")
        username, token = decoded.split(":", 1)
        try:
            token = Token.objects.get(key=token)
            if token.user.username == username:
                return token.user
        except:
            pass


def get_token(request):
    """The same as validate_token, but return the token object to check the
    associated user.

    Arguments:
    ==========
    request (requests.Request)    : the Request object to inspect
    """
    # Coming from HTTP, look for authorization as bearer token
    token = request.META.get("HTTP_AUTHORIZATION")

    if token:
        try:
            return Token.objects.get(key=token.replace("BEARER", "").strip())
        except Token.DoesNotExist:
            pass

    # Next attempt - try to get token via user session
    elif request.user.is_authenticated and not request.user.is_anonymous:
        try:
            return Token.objects.get(user=request.user)
        except Token.DoesNotExist:
            pass


def get_challenge(request, repository, scopes=["pull", "push"]):
    """Given an unauthenticated request, return a challenge in
    the Www-Authenticate header

    Arguments:
    ==========
    request (requests.Request): the Request object to inspect
    repository (str)          : the repository name
    scopes (list)             : list of scopes to return
    """
    DOMAIN_NAME = get_server(request)
    if not isinstance(scopes, list):
        scopes = [scopes]
    auth_server = settings.AUTH_SERVER or "%s/auth/token" % DOMAIN_NAME
    return 'realm="%s",service="%s",scope="repository:%s:%s"' % (
        auth_server,
        DOMAIN_NAME,
        repository,
        ",".join(scopes),
    )



def return_errors():

    {
        "errors": [
            {
                "code": "<error identifier>",
                "message": "<message describing condition>",
                "detail": "<unstructured>",
            },
            ...,
        ]
    }



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer
from django.http.response import Http404
from django.views.decorators.cache import never_cache

from django_oci.models import Repository, get_image_by_tag
from django_oci import settings
from .parsers import ManifestRenderer
from django_oci.auth import is_authenticated

from ratelimit.decorators import ratelimit
from django.utils.decorators import method_decorator


class ImageTags(APIView):
    """Return a list of tags for an image."""

    permission_classes = []
    allowed_methods = ("GET",)

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """GET /v2/<name>/tags/list. We don't require authentication to list tags,
        unless the repository is private.
        """
        name = kwargs.get("name")
        number = request.GET.get("n")
        last = request.GET.get("last")

        try:
            repository = Repository.objects.get(name=name)
        except Repository.DoesNotExist:
            raise Http404

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(request, repository)
        if not allow_continue:
            return response

        tags = [
            x
            for x in list(repository.image_set.values_list("tag__name", flat=True))
            if x
        ]

        # Tags must be sorted in lexical order
        tags.sort()

        # Number must be an integer if defined
        if number:
            number = int(number)

        # if last, <tagname> not included in the results, but up to <int> tags after <tagname> will be returned.
        if last and number:
            try:
                start = tags.index(last)
            except IndexError:
                start = 0
            tags = tags[start:number]

        elif number:
            tags = tags[:number]

        # Ensure tags sorted in lexical order
        data = {"name": repository.name, "tags": sorted(tags)}
        return Response(status=200, data=data)


class ImageManifest(APIView):
    """An Image Manifest holds the configuration and metadata about an image
    GET: is to retrieve an existing image manifest
    PUT: is to push a manifest
    """

    renderer_classes = [ManifestRenderer, JSONRenderer]
    permission_classes = []
    allowed_methods = (
        "GET",
        "PUT",
        "DELETE",
    )

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="DELETE",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def delete(self, request, *args, **kwargs):
        """DELETE /v2/<name>/manifests/<tag>"""

        # A registry must globally disable or enable both
        if settings.DISABLE_TAG_MANIFEST_DELETE:
            return Response(status=405)

        name = kwargs.get("name")
        reference = kwargs.get("reference")
        tag = kwargs.get("tag")

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(
            request, name, must_be_owner=True
        )
        if not allow_continue:
            return response

        # Retrieve the image, return of None indicates not found
        image = get_image_by_tag(name, reference=reference, tag=tag, create=False)
        if not image:
            raise Http404

        # Delete the image tag
        if tag:
            tag = image.tag_set.filter(name=tag)
            tag.delete()

        # Delete a manifest
        elif reference:
            image.delete()

        # Upon success, the registry MUST respond with a 202 Accepted code.
        return Response(status=202)

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="PUT",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def put(self, request, *args, **kwargs):
        """PUT /v2/<name>/manifests/<reference>
        https://github.com/opencontainers/distribution-spec/blob/master/spec.md#pushing-manifests
        """
        # We likely can default to the v1 manifest, unless otherwise specified
        # This isn't used or checked for the time being
        # application/vnd.oci.image.manifest.v1+json
        _ = request.META.get("CONTENT_TYPE", settings.IMAGE_MANIFEST_CONTENT_TYPE)

        name = kwargs.get("name")
        reference = kwargs.get("reference")
        tag = kwargs.get("tag")

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(
            request, name, must_be_owner=True
        )
        if not allow_continue:
            return response

        # Reference must be sha256 digest
        if reference and not reference.startswith("sha256:"):
            return Response(status=400)

        # Also provide the body in case we have a tag
        image = get_image_by_tag(name, reference, tag, create=True, body=request.body)

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(
            request, image.repository, must_be_owner=True
        )
        if not allow_continue:
            return response

        return Response(status=201, headers={"Location": image.get_manifest_url()})

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """GET /v2/<name>/manifests/<reference>"""

        name = kwargs.get("name")
        reference = kwargs.get("reference")
        tag = kwargs.get("tag")

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(request, name)
        if not allow_continue:
            return response

        image = get_image_by_tag(name, tag=tag, reference=reference)

        # If the manifest is not found in the registry, the response code MUST be 404 Not Found.
        if not image:
            raise Http404
        return Response(image.manifest, status=200)



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from django.core.files.base import ContentFile
from django.core.files.uploadedfile import UploadedFile
from django_oci.settings import MEDIA_ROOT
from django.db import models
from datetime import timezone

import os
import hashlib


class ChunkedUpload(models.Model):
    """We can use an abstract class to interact with a chunked upload without
    saving anything to the database.
    """

    session_id = models.CharField()
    file = models.FileField(
        max_length=255, upload_to=os.path.join(MEDIA_ROOT, "sessions")
    )
    offset = models.BigIntegerField(default=0)

    @property
    def filename(self):
        return os.path.basename(self.session_id)

    @property
    def expires_on(self):
        return self.created_on + 10000

    @property
    def expired(self):
        return self.expires_on <= timezone.now()

    @property
    def sha256(self):
        if getattr(self, "_md5", None) is None:
            hasher = hashlib.sha256()
            for chunk in self.file.chunks():
                hasher.update(chunk)
            self._sha256 = hasher.hexdigest()
        return self._sha256

    def delete(self, delete_file=True, *args, **kwargs):
        if self.file:
            storage, path = self.file.storage, self.file.path
        super(ChunkedUpload, self).delete(*args, **kwargs)
        if self.file and delete_file:
            storage.delete(path)

    def __str__(self):
        return "<%s: session_id: %s - bytes: %s>" % (
            self.filename,
            self.session_id,
            self.offset,
        )

    def write_chunk(self, chunk, chunk_start):
        """Append a chunk to the file, or write the file if it doesn't exist yet.
        This is done to a temporary storage location in images/sessions until
        the blob is finalized.
        """
        self.file.close()

        # If it's the first chunk, we need to instantiate the file
        if chunk_start == 0:
            self.file.save(name=self.filename, content=ContentFile(""), save=False)

        # We should start writing at next index, not over a previously written one
        if self.file.size == 0 and chunk_start != 0:
            return 415

        # If a chunk is uploaded out of order, the registry MUST respond with a 416 Requested Range Not Satisfiable code.
        elif chunk_start != 0 and self.file.size != chunk_start:
            return 416

        # Write chunk (mode = append+binary)
        with open(self.file.path, mode="ab") as fd:
            fd.write(chunk)

        # Update the offset
        self.offset += len(chunk)
        self._sha256 = None  # Clear cached hash digest
        self.file.close()  # Flush
        return 202

    def get_uploaded_file(self):
        self.file.close()
        self.file.open(mode="rb")  # mode = read+binary
        return UploadedFile(file=self.file, name=self.filename, size=self.offset)

    class Meta:
        abstract = True



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from django.core.files.storage import FileSystemStorage
from django_oci import settings
from django.urls import reverse
from django.db import models
from django.contrib.auth.models import User
from django.middleware import cache

import hashlib
import json
import os
import re


PRIVACY_CHOICES = (
    (False, "Public (The collection will be accessible by anyone)"),
    (True, "Private (The collection will be not listed.)"),
)


def get_privacy_default():
    return settings.PRIVATE_ONLY


class OverwriteStorage(FileSystemStorage):
    def get_available_name(self, name, **kwargs):
        """Define a new filesystem storage so that it's okay to overwrite an
        existing filename.
        """
        if self.exists(name):
            os.remove(name)
        return name

    def get_valid_name(self, name, **kwargs):
        """The default function will replace the : in the filename (removing it)
        but we want to allow it.
        """
        name = str(name).strip().replace(" ", "_")
        return re.sub(r"(?u)[^-\w.:]", "", name)


def calculate_digest(body):
    """Calculate the sha256 sum for some body (bytes)"""
    hasher = hashlib.sha256()
    hasher.update(body)
    return hasher.hexdigest()


def get_upload_folder(instance, filename):
    """a helper function to upload a blob to local storage"""
    repository_name = instance.repository.name
    blobs_home = os.path.join(settings.MEDIA_ROOT, "blobs", repository_name)
    if not os.path.exists(blobs_home):
        os.makedirs(blobs_home)
    filename = os.path.join(blobs_home, filename)
    return filename


def get_image_by_tag(name, reference, tag, create=False, body=None):
    """given the name of a repository and a reference, look up the image
    based on the reference. By default we use the reference to look for
    a tag or digest. A return of None indicates that the image is not found,
    and the view should deal with this (e.g., create the image) or raise
    Http404.

    Parameters
    ==========
    name (str): the name of the repository to lookup
    reference (str): an image version string
    tag (str): a tag that doesn't match as a version string
    create (bool): if does not exist, create the image (new manifest push)
    body (bytes): if we need to create, we must have a digest from the body
    """
    # Ensure the repository exists
    try:
        repository = Repository.objects.get(name=name)
    except Repository.DoesNotExist:
        return None

    # reference can be a tag (more likely) or digest
    image = None
    if tag:
        try:
            image = repository.image_set.get(tag__name=tag)
        except Image.DoesNotExist:
            pass

    elif reference:
        try:
            image = repository.image_set.get(version=reference)
        except Image.DoesNotExist:
            pass

    if not image and create:
        if not reference and body:
            reference = "sha256:%s" % calculate_digest(body)
        image, _ = Image.objects.get_or_create(
            repository=repository, version=reference, manifest=body
        )
        if tag:
            tag, _ = Tag.objects.get_or_create(image=image, name=tag)
            tag.image = image
            tag.save()

        # This saves annotations and layer (blob) associations
        image.update_manifest(body)

    return image


class Repository(models.Model):

    name = models.CharField(
        max_length=500,  # name of repository, e.g., username/reponame
        unique=True,
        blank=False,
        null=False,
    )

    add_date = models.DateTimeField("date added", auto_now_add=True)
    modify_date = models.DateTimeField("date modified", auto_now=True)
    owners = models.ManyToManyField(
        User,
        blank=True,
        default=None,
        related_name="container_collection_owners",
        related_query_name="owners",
    )
    contributors = models.ManyToManyField(
        User,
        related_name="container_collection_contributors",
        related_query_name="contributor",
        blank=True,
        help_text="users with edit permission to the collection",
        verbose_name="Contributors",
    )

    # By default, collections are public
    private = models.BooleanField(
        choices=PRIVACY_CHOICES,
        default=get_privacy_default,
        verbose_name="Accessibility",
    )

    def has_view_permission(self, user):
        return user in self.owners.all() or user in self.contributors.all()

    def get_absolute_url(self):
        return reverse("repository_details", args=[str(self.id)])

    def __str__(self):
        return self.get_uri()

    def __unicode__(self):
        return self.get_uri()

    def get_uri(self):
        return self.name

    def get_label(self):
        return "repository"

    class Meta:
        app_label = "django_oci"


class Blob(models.Model):
    """a blob, which can be a binary or archive to be extracted."""

    add_date = models.DateTimeField("date added", auto_now_add=True)
    modify_date = models.DateTimeField("date modified", auto_now=True)
    content_type = models.CharField(max_length=250, null=False)
    digest = models.CharField(max_length=250, null=True, blank=True)
    datafile = models.FileField(
        upload_to=get_upload_folder, max_length=255, storage=OverwriteStorage()
    )
    remotefile = models.CharField(max_length=500, null=True, blank=True)

    # When a repository is deleted, so are the blobs
    repository = models.ForeignKey(
        Repository,
        null=False,
        blank=False,
        on_delete=models.CASCADE,
    )

    def get_label(self):
        return "blob"

    def get_download_url(self):
        """Blobs are loosely associated with repositories."""
        if self.remotefile is not None:
            return self.remotefile
        return settings.DOMAIN_URL.strip("/") + reverse(
            "django_oci:blob_download",
            kwargs={"digest": self.digest, "name": self.repository.name},
        )

    def create_upload_session(self):
        """A function to create an upload session for a particular blob.
        The version variable will be set with a session id.
        """
        # Get the django oci upload cache, and generate an expiring session upload id
        filecache = cache.caches["django_oci_upload"]

        # Expires in default 10 seconds
        filecache.set(self.session_id, 1, timeout=settings.SESSION_EXPIRES_SECONDS)
        return reverse("django_oci:blob_upload", kwargs={"session_id": self.session_id})

    @property
    def session_id(self):
        return "put/%s/%s" % (self.id, self.digest)

    def get_abspath(self):
        return os.path.join(settings.MEDIA_ROOT, self.datafile.name)

    class Meta:
        app_label = "django_oci"
        unique_together = (
            (
                "repository",
                "digest",
            ),
        )


class Image(models.Model):
    """An image (manifest) holds a set of layers (blobs) for a repository.
    Blobs can be shared between manifests, and are deleted if they are
    no longer referenced.
    """

    add_date = models.DateTimeField("date manifest added", auto_now=True)

    # When a repository is deleted, so are the manifests
    repository = models.ForeignKey(
        Repository,
        null=False,
        blank=False,
        on_delete=models.CASCADE,
    )

    # Blobs are added at the creation of the manifest (can be shared based on hash)
    blobs = models.ManyToManyField(Blob)

    # The text of the manifest (added at the end)
    manifest = models.BinaryField(null=False, blank=False, default=b"{}")

    # The version (digest) of the manifest
    version = models.CharField(max_length=250, null=True, blank=True)

    # Manifest functions to get, save, and return download url
    def get_manifest(self):
        return self.manifest

    def add_blob(self, digest):
        """A helper function to lookup and add a blob to an image. If the blob
        is already added, no harm done.
        """
        if digest:
            try:
                blob = Blob.objects.get(digest=digest)
                self.blobs.add(blob)
            except Blob.DoesNotExist:
                pass

    def remove_blob(self, digest):
        """We can only remove a blob if it is no longer linked to any images
        for the repository. This should be called after new blobs are parsed
        and added via the manifest.
        """
        if digest:
            try:
                blob = Blob.objects.get(digest=digest, repository=self.repository)
                if blob.image_set.count() == 0:
                    blob.delete()
            except Blob.DoesNotExist:
                pass

    def update_blob_links(self, manifest):
        # Keep a list of blobs that we will remove
        current_blobs = set()

        # The configuration blob, and then all blob layers
        config_digest = manifest.get("config", {}).get("digest")
        current_blobs.add(config_digest)
        for layer in manifest.get("layers", []):
            current_blobs.add(layer.get("digest"))

        # Remove unlinked blobs
        unlinked_blobs = [x for x in self.blobs.all() if x.digest not in current_blobs]

        # Add all current blobs not already present, remove unlinked
        [self.add_blob(digest) for digest in current_blobs]
        [self.remove_blob(x) for x in unlinked_blobs]

    def update_annotations(self, manifest):
        # Just delete all previous annotations
        self.annotation_set.all().delete()
        for key, value in manifest.get("annotations", {}).items():
            annotation, created = Annotation.objects.get_or_create(key=key, image=self)
            annotation.value = value
            annotation.save()

    def update_manifest(self, manifest):
        """Loading a manifest (after save) means creating an association between blobs and
        annotations
        """
        # Load a derivation to get blob links and annotations
        if not isinstance(manifest, str):
            manifest = manifest.decode("utf-8")
        if not isinstance(manifest, dict):
            manifest = json.loads(manifest)
        self.update_blob_links(manifest)
        self.update_annotations(manifest)
        self.save()

    def get_manifest_url(self):
        if self.version:
            return reverse(
                "django_oci:image_manifest",
                kwargs={"name": self.repository.name, "reference": self.version},
            )
        return reverse(
            "django_oci:image_manifest",
            kwargs={"name": self.repository.name, "tag": self.tag_set.first().name},
        )

    # A container only gets a version when fit's frozen, otherwise known by tag
    def get_uri(self):
        return self.repository.name

    # Return an image file path
    def get_image_path(self):
        if self.image not in [None, ""]:
            return self.image.datafile.path
        return None

    def get_download_url(self):
        if self.image not in [None, ""]:
            return self.image.datafile.file
        return None

    def get_label(self):
        return "image"

    def __str__(self):
        return self.get_uri()

    def __unicode__(self):
        return self.get_uri()

    class Meta:
        app_label = "django_oci"
        unique_together = (
            (
                "repository",
                "version",
            ),
        )


class Tag(models.Model):
    """A tag is a reference for one or more manifests"""

    name = models.CharField(max_length=250, null=False, blank=False)
    image = models.ForeignKey(
        Image,
        null=False,
        blank=False,
        # When a manifest is deleted, any associated tags are too
        on_delete=models.CASCADE,
    )

    def __str__(self):
        return "<tag:%s>" % self.name


class Annotation(models.Model):
    """An annotation is a key/value pair to describe an image.
    We will want to parse these from an image manifest (eventually)
    """

    key = models.CharField(max_length=250, null=False, blank=False)
    value = models.CharField(max_length=250, null=False, blank=False)
    image = models.ForeignKey(Image, on_delete=models.CASCADE)

    def __str__(self):
        return "%s:%s" % (self.key, self.value)

    def __unicode__(self):
        return "%s:%s" % (self.key, self.value)

    def get_label(self):
        return "annotation"

    class Meta:
        app_label = "django_oci"
        unique_together = (("key", "image"),)



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import logging
import re

logger = logging.getLogger(__name__)


# Regular expressions to parse registry, collection, repo, tag and version
_docker_uri = re.compile(
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    "(?P<collection>(?:[^:@/]+/)+)?"
    "(?P<repo>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<version>.+))?"
    "$"
)

# Reduced to match registry:port/repo or registry.com/repo
_reduced_uri = re.compile(
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    "(?P<repo>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<version>.+))?"
    "$"
    "(?P<collection>.)?"
)

# Default
_default_uri = re.compile(
    "(?:(?P<registry>[^/@]+)/)?"
    "(?P<collection>(?:[^:@/]+/)+)"
    "(?P<repo>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<version>.+))?"
    "$"
)


def set_default(item, default, use_default):
    """if an item provided is None and boolean use_default is set to True,
    return the default. Otherwise, return the item.
    """
    if item is None and use_default:
        return default
    return item


def get_server(request):
    """Given a request, parse it to determine the server name and using http/https"""
    scheme = request.is_secure() and "https" or "http"
    return f"{scheme}://{request.get_host()}"


def parse_content_range(content_range):
    """Given a content range, match based on regular expression and return
    parsed start, end (both int)
    """
    # Ensure range matches regular expression
    if not re.search("^[0-9]+-[0-9]+$", content_range):
        raise ValueError

    # Parse the content range into numbers
    return [int(x.strip()) for x in content_range.strip().split("-")]


def parse_image_name(
    image_name,
    tag=None,
    version=None,
    defaults=True,
    ext="sif",
    default_collection="library",
    default_tag="latest",
    base=None,
    lowercase=True,
):

    """return a collection and repo name and tag
    for an image file.

    Parameters
    =========
    image_name: a user provided string indicating a collection,
                image, and optionally a tag.
    tag: optionally specify tag as its own argument
         over-rides parsed image tag
    defaults: use defaults "latest" for tag and "library"
              for collection.
    base: if defined, remove from image_name, appropriate if the
          user gave a registry url base that isn't part of namespace.
    lowercase: turn entire URI to lowercase (default is True)
    """

    # Save the original string
    original = image_name

    if base is not None:
        image_name = image_name.replace(base, "").strip("/")

    # If a file is provided, remove extension
    image_name = re.sub("[.](img|simg|sif)", "", image_name)

    # Parse the provided name
    uri_regexes = [_reduced_uri, _default_uri, _docker_uri]

    for r in uri_regexes:
        match = r.match(image_name)
        if match:
            break

    if not match:
        logger.warning('Could not parse image "%s"! Exiting.' % image_name)
        return

    # Get matches
    registry = match.group("registry")
    collection = match.group("collection")
    repo_name = match.group("repo")
    repo_tag = tag or match.group("tag")
    version = version or match.group("version")

    # A repo_name is required
    assert repo_name

    # If a collection isn't provided
    collection = set_default(collection, default_collection, defaults)
    repo_tag = set_default(repo_tag, default_tag, defaults)

    # The collection, name must be all lowercase
    if lowercase:
        collection = collection.lower().rstrip("/")
        repo_name = repo_name.lower()
        repo_tag = repo_tag.lower()
    else:
        collection = collection.rstrip("/")

    if version is not None:
        version = version.lower()

    # Piece together the uri base
    if registry is None:
        uri = "%s/%s" % (collection, repo_name)
    else:
        uri = "%s/%s/%s" % (registry, collection, repo_name)

    url = uri

    # Tag is defined
    if repo_tag is not None:
        uri = "%s:%s" % (url, repo_tag)

    # Version is defined
    storage_version = None
    if version is not None:
        uri = "%s@%s" % (uri, version)
        storage_version = "%s.%s" % (uri, ext)

    # A second storage URI honors the tag (:) separator

    storage = "%s.%s" % (uri, ext)
    return {
        "collection": collection,
        "original": original,
        "registry": registry,
        "image": repo_name,
        "url": url,
        "tag": repo_tag,
        "version": version,
        "storage": storage_version or storage,
        "uri": uri,
    }



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from .models import Image
from rest_framework.authtoken.models import Token
from django.contrib.auth import get_user_model

UserModel = get_user_model()


@receiver(post_delete, sender=Image)
def delete_blobs(sender, instance, **kwargs):
    print("Delete image signal running.")

    for blob in instance.blobs.all():
        if hasattr(blob, "datafile"):
            if blob.image_set.count() == 0:
                print("Deleting %s, no longer used." % blob.datafile)
                blob.datafile.delete()
                blob.delete()


@receiver(post_save, sender=UserModel)
def create_auth_token(sender, instance=None, created=False, **kwargs):
    """Create a token for the user when the user is created (with oAuth2)

    1. Assign user a token
    2. Assign user to default group

    Create a Profile instance for all newly created User instances. We only
    run on user creation to avoid having to check for existence on each call
    to User.save.
    """
    if created:
        Token.objects.create(user=instance)



from .base import APIVersionCheck
from .blobs import BlobUpload, BlobDownload
from .image import ImageManifest, ImageTags
from .auth import GetAuthToken

# Load storage on application init
from django_oci.storage import get_storage

storage = get_storage()



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from rest_framework import renderers
from rest_framework.negotiation import BaseContentNegotiation


class IgnoreClientContentNegotiation(BaseContentNegotiation):
    """Important! This class is not used, however if you want to validate
    (or allow) any media type (e.g., the server doesn't return 406) you
    can set this as the default content negotiation for any view.

    content_negotiation_class = IgnoreClientContentNegotiation
    """

    def select_parser(self, request, parsers):
        """Return the first parser (usually application/json)"""
        return parsers[0]

    def select_renderer(self, request, renderers, format_suffix):
        """Return the first renderer (usually application/json)"""
        return (renderers[0], renderers[0].media_type)


class ManifestRenderer(renderers.BaseRenderer):
    """A ManifestRenderer is provided to the ImageManifest views, which need
    to accept a content type for the image manifest. Without this
    renderer, Django will return 406
    """

    media_type = "application/vnd.oci.image.manifest.v1+json"
    format = None
    charset = None
    render_style = "binary"

    def render(self, data, media_type=None, renderer_context=None):
        return data



# Generated by Django 3.1.2 on 2020-10-28 16:27

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_oci.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Blob',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('add_date', models.DateTimeField(auto_now_add=True, verbose_name='date added')),
                ('modify_date', models.DateTimeField(auto_now=True, verbose_name='date modified')),
                ('content_type', models.CharField(max_length=250)),
                ('digest', models.CharField(blank=True, max_length=250, null=True)),
                ('datafile', models.FileField(max_length=255, storage=django_oci.models.OverwriteStorage(), upload_to=django_oci.models.get_upload_folder)),
                ('remotefile', models.CharField(blank=True, max_length=500, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('add_date', models.DateTimeField(auto_now=True, verbose_name='date manifest added')),
                ('manifest', models.BinaryField(default=b'{}')),
                ('version', models.CharField(blank=True, max_length=250, null=True)),
                ('blobs', models.ManyToManyField(to='django_oci.Blob')),
            ],
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('image', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_oci.image')),
            ],
        ),
        migrations.CreateModel(
            name='Repository',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500, unique=True)),
                ('add_date', models.DateTimeField(auto_now_add=True, verbose_name='date added')),
                ('modify_date', models.DateTimeField(auto_now=True, verbose_name='date modified')),
                ('private', models.BooleanField(choices=[(False, 'Public (The collection will be accessible by anyone)'), (True, 'Private (The collection will be not listed.)')], default=django_oci.models.get_privacy_default, verbose_name='Accessibility')),
                ('contributors', models.ManyToManyField(blank=True, help_text='users with edit permission to the collection', related_name='container_collection_contributors', related_query_name='contributor', to=settings.AUTH_USER_MODEL, verbose_name='Contributors')),
                ('owners', models.ManyToManyField(blank=True, default=None, related_name='container_collection_owners', related_query_name='owners', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='image',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_oci.repository'),
        ),
        migrations.AddField(
            model_name='blob',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_oci.repository'),
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together={('repository', 'version')},
        ),
        migrations.AlterUniqueTogether(
            name='blob',
            unique_together={('repository', 'digest')},
        ),
        migrations.CreateModel(
            name='Annotation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=250)),
                ('value', models.CharField(max_length=250)),
                ('image', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_oci.image')),
            ],
            options={
                'unique_together': {('key', 'image')},
            },
        ),
    ]



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.decorators import authentication_classes, permission_classes
from django.views.decorators.cache import never_cache

from django_oci import settings
from django.http import HttpResponseForbidden

from django_oci.utils import get_server
from django_oci.auth import get_user, generate_jwt

import re


@authentication_classes([])
@permission_classes([])
class GetAuthToken(APIView):
    """Given a GET request for a token, validate and return it."""

    permission_classes = []
    allowed_methods = ("GET",)

    @never_cache
    def get(self, request, *args, **kwargs):
        """GET /auth/token"""
        print("GET /auth/token")
        user = get_user(request)

        # No token provided matching a user, no go
        if not user:
            return HttpResponseForbidden()

        # Formalate the jwt token response, with a unique id
        _ = request.GET.get("service")
        scope = request.GET.get("scope")

        # The scope should include the repository name, and permissions
        # "repository:vanessa/container:pull,push"
        match = re.match("repository:(?P<repository>.+):(?P<scope>.+)", scope)
        repository, scope = match.groups()
        scope = scope.split(",")

        # Generate domain name for auth server
        DOMAIN_NAME = get_server(request)
        auth_server = settings.AUTH_SERVER or "%s/auth/token" % DOMAIN_NAME

        # Generate the token data, a dict with token, expires_in, and issued_at
        data = generate_jwt(
            username=user.username,
            scope=scope,
            realm=auth_server,
            repository=repository,
        )
        return Response(status=200, data=data)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from django.views.generic.base import TemplateView
from chunked_upload.views import ChunkedUploadView, ChunkedUploadCompleteView
from django_oci.models import MyChunkedUpload


class ChunkedUploadDemo(TemplateView):
    template_name = "django_oci/chunked_upload.html"


class MyChunkedUploadView(ChunkedUploadView):

    model = MyChunkedUpload
    field_name = "the_file"

    def check_permissions(self, request):
        # Allow non authenticated users to make uploads
        print("django_oci:check_permissions")
        pass


class MyChunkedUploadCompleteView(ChunkedUploadCompleteView):

    model = MyChunkedUpload

    def check_permissions(self, request):
        # Allow non authenticated users to make uploads
        print("django_oci:check_permissions")
        pass

    def on_completion(self, uploaded_file, request):
        # Do something with the uploaded file. E.g.:
        # * Store the uploaded file on another model:
        # SomeModel.objects.create(user=request.user, file=uploaded_file)
        # * Pass it as an argument to a function:
        # function_that_process_file(uploaded_file)
        print("django_oci:on_completion")
        pass

    def get_response_data(self, chunked_upload, request):
        print("django_oci:get_response_data")
        return {
            "message": (
                "You successfully uploaded '%s' (%s bytes)!"
                % (chunked_upload.filename, chunked_upload.offset)
            )
        }



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

Settings are namespaced to the DJANGO_OCI namespace. Any of the following
below can be overridden in the default django settings, e.g.:

DJANGO_OCI = {
    'SPEC_VERSION': "v2",
    'PRIVATE_ONLY': False
}

"""

from django.conf import settings
import logging
import uuid
import os

logger = logging.getLogger(__name__)

# Default views to Authenticate
authenticated_views = [
    "django_oci.views.blobs.BlobUpload",
    "django_oci.views.blobs.BlobDownload",
    "django_oci.views.image.ImageTags",
    "django_oci.views.image.ImageManifest",
]

DEFAULTS = {
    # Url base prefix
    "URL_PREFIX": "v2",
    # Disable registry authentication
    "DISABLE_AUTHENTICATION": False,
    # Version of distribution spec
    "SPEC_VERSION": "1",
    # Repository permissions
    "PRIVATE_ONLY": False,
    # Allowed content types to upload as layers
    "CONTENT_TYPES": ["application/octet-stream"],
    # Image Manifest content type
    "IMAGE_MANIFEST_CONTENT_TYPE": "application/vnd.oci.image.manifest.v1+json",
    # Storage backend
    "STORAGE_BACKEND": "filesystem",
    # Domain used in templates, api prefix
    "DOMAIN_URL": "http://127.0.0.1:8000",
    # Media root (if saving images on filesystem
    "MEDIA_ROOT": "images",
    # Set a cache directory, otherwise defaults to MEDIA_ROOT + /cache
    "CACHE_DIR": None,
    # The number of seconds a session (upload request) is valid (10 minutes)
    "SESSION_EXPIRES_SECONDS": 600,
    # The number of seconds a token is valid (10 minutes)
    "TOKEN_EXPIRES_SECONDS": 600,
    # Disable deletion of an image by tag or manifest (default is not disabled)
    "DISABLE_TAG_MANIFEST_DELETE": False,
    # Default content type is application/octet-stream
    "DEFAULT_CONTENT_TYPE": "application/octet-stream",
    # Default views to put under authentication, given that DISABLE_AUTHENTICTION is False
    "AUTHENTICATED_VIEWS": authenticated_views,
    # If you have a custom authentication server to generate tokens (defaults to /registry/auth/token
    "AUTHENTICATION_SERVER": None,
    # jwt encoding secret: set server wide or generated on the fly
    "JWT_SERVER_SECRET": str(uuid.uuid4()),
    # View rate limit, defaults to 100/1day using django-ratelimit based on ipaddress
    "VIEW_RATE_LIMIT": "100/1d",
    # Given that someone goes over, are they blocked for a period?
    "VIEW_RATE_LIMIT_BLOCK": True,
}

# The user can define a section for DJANGO_OCI in settings
oci = getattr(settings, "DJANGO_OCI", DEFAULTS)

# Allows for import of variables above from django_oci.settings
URL_PREFIX = oci.get("URL_PREFIX", DEFAULTS["URL_PREFIX"])
DISABLE_AUTHENTICATION = oci.get(
    "DISABLE_AUTHENTICATION", DEFAULTS["DISABLE_AUTHENTICATION"]
)
AUTH_SERVER = oci.get("AUTHENTICATION_SERVER", DEFAULTS["AUTHENTICATION_SERVER"])
SPEC_VERSION = oci.get("SPEC_VERSION", DEFAULTS["SPEC_VERSION"])
PRIVATE_ONLY = oci.get("PRIVATE_ONLY", DEFAULTS["PRIVATE_ONLY"])
CONTENT_TYPES = oci.get("CONTENT_TYPES", DEFAULTS["CONTENT_TYPES"])
JWT_SERVER_SECRET = oci.get("JWT_SERVER_SECRET", DEFAULTS["JWT_SERVER_SECRET"])
STORAGE_BACKEND = oci.get("STORAGE_BACKEND", DEFAULTS["STORAGE_BACKEND"])
DOMAIN_URL = oci.get("DOMAIN_URL", DEFAULTS["DOMAIN_URL"])
MEDIA_ROOT = oci.get("MEDIA_ROOT", DEFAULTS["MEDIA_ROOT"])
CACHE_DIR = oci.get("CACHE_DIR", DEFAULTS["CACHE_DIR"])
DEFAULT_CONTENT_TYPE = oci.get("DEFAULT_CONTENT_TYPE", DEFAULTS["DEFAULT_CONTENT_TYPE"])
IMAGE_MANIFEST_CONTENT_TYPE = oci.get(
    "IMAGE_MANIFEST_CONTENT_TYPE", DEFAULTS["IMAGE_MANIFEST_CONTENT_TYPE"]
)
DISABLE_TAG_MANIFEST_DELETE = oci.get(
    "DISABLE_TAG_MANIFEST_DELETE", DEFAULTS["DISABLE_TAG_MANIFEST_DELETE"]
)
TOKEN_EXPIRES_SECONDS = oci.get(
    "TOKEN_EXPIRES_SECONDS", DEFAULTS["TOKEN_EXPIRES_SECONDS"]
)
SESSION_EXPIRES_SECONDS = oci.get(
    "SESSION_EXPIRES_SECONDS", DEFAULTS["SESSION_EXPIRES_SECONDS"]
)
AUTHENTICATED_VIEWS = oci.get("AUTHENTICATED_VIEWS", DEFAULTS["AUTHENTICATED_VIEWS"])

# Rate Limits
VIEW_RATE_LIMIT = oci.get("VIEW_RATE_LIMIT", DEFAULTS["VIEW_RATE_LIMIT"])
VIEW_RATE_LIMIT_BLOCK = oci.get(
    "VIEW_RATE_LIMIT_BLOCK", DEFAULTS["VIEW_RATE_LIMIT_BLOCK"]
)


# Set filesystem cache, also adding to middleware
CACHES = getattr(settings, "CACHES", {})
MIDDLEWARE = getattr(settings, "MIDDLEWARE", [])

for entry in [
    "django.middleware.cache.UpdateCacheMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.cache.FetchFromCacheMiddleware",
]:
    if entry not in MIDDLEWARE:
        MIDDLEWARE.append(entry)

# Create a filesystem cache for temporary upload sessions
cache = CACHE_DIR or os.path.join(MEDIA_ROOT, "cache")
if not os.path.exists(cache):
    logger.debug(f"Creating cache directory {cache}")
    os.makedirs(cache)

CACHES.update(
    {
        "django_oci_upload": {
            "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
            "LOCATION": cache,
        }
    }
)



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from rest_framework.views import APIView
from rest_framework.response import Response
from django.views.decorators.cache import never_cache

from django_oci.models import Blob, Repository
from django_oci import settings
from django_oci.storage import storage
from django.middleware import cache

from django_oci.utils import parse_content_range
from django_oci.auth import is_authenticated

from ratelimit.decorators import ratelimit
from django.utils.decorators import method_decorator


class BlobDownload(APIView):
    """Given a GET request for a blob, stream the blob."""

    permission_classes = []
    allowed_methods = (
        "GET",
        "DELETE",
    )

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """POST /v2/<name>/blobs/<digest>"""
        # the name is only used to validate the user has permission to upload
        name = kwargs.get("name")
        digest = kwargs.get("digest")

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(request, name)
        if not allow_continue:
            return response

        return storage.download_blob(name, digest)

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="DELETE",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def delete(self, request, *args, **kwargs):
        """DELETE /v2/<name>/blobs/<digest>"""
        name = kwargs.get("name")
        digest = kwargs.get("digest")

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(
            request, name, must_be_owner=True
        )
        if not allow_continue:
            return response

        return storage.delete_blob(name, digest)


class BlobUpload(APIView):
    """An image push will receive a request to push, authenticate the user,
    and return an upload url (url is /v2/<name>/blobs/uploads/)
    """

    permission_classes = []
    allowed_methods = (
        "POST",
        "PUT",
        "PATCH",
    )

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="PUT",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def put(self, request, *args, **kwargs):
        """PUT /v2/<name>/blobs/uploads/
        A put request can happen in two scenarios. 1. after a POST request,
        and must include a session_id. The session id is created via the file
        system cache, and each one includes the request type, image id
        (associated with a tag), repository, and a randomly generated uuid.
        The upload will not continue if any required metadata is missing or
        the identifier is already expired. The second case is after several
        PATCH requests to upload chunks of a blob. The final chunk might be
        provided this case.
        """
        # These header attributes are shared by both scenarios
        session_id = kwargs.get("session_id")
        digest = request.GET.get("digest")
        content_length = int(request.META.get("CONTENT_LENGTH", 0))
        content_type = request.META.get("CONTENT_TYPE", settings.DEFAULT_CONTENT_TYPE)

        # Presence of content range distinguishes chunked upload from single PUT
        # A final PUT request may not have a content_range if no chunk to upload
        content_range = request.META.get("HTTP_CONTENT_RANGE")

        if not session_id or not digest or not content_type:
            return Response(status=400)

        # Confirm that content length (body) == header value, otherwise bad request
        if len(request.body) != content_length:
            return Response(status=400)

        # Get the session id, if it has not expired
        filecache = cache.caches["django_oci_upload"]
        if not filecache.get(session_id):
            return Response(status=400)

        # Ensure it cannot be used again
        filecache.set(session_id, None, timeout=0)

        # Break apart into blob id, and session uuid (version)
        _, blob_id, version = session_id.split("/")
        try:
            blob = Blob.objects.get(id=blob_id, digest=version)
        except Blob.DoesNotExist:
            return Response(status=404)

        # If allow_continue False, return response
        allow_continue, response, _ = is_authenticated(
            request, blob.repository, must_be_owner=True
        )
        print(allow_continue)
        print(response)
        if not allow_continue:
            return response

        if not content_range and request.body:

            # Now process the PUT request to the file! Provide the blob to update
            return storage.create_blob(
                blob=blob,
                body=request.body,
                digest=digest,
                content_type=content_type,
            )

        # Scenario 2: a PUT to end a chunked upload session, no final chunk
        elif not request.body:
            return storage.finish_blob(
                blob=blob,
                digest=digest,
            )

        # Scenario 3: a PUT to end a chunked upload session with a final chunk
        # First ensure upload session cannot be used again
        filecache.set(session_id, None, timeout=0)

        # Parse the start and end for the chunk to write
        try:
            content_start, content_end = parse_content_range(content_range)
        except ValueError:
            return Response(status=400)

        # Write the final chunk and finish the session
        status_code = storage.write_chunk(
            blob=blob,
            content_start=content_start,
            content_end=content_end,
            body=request.body,
        )

        # If it's already existing, return Accepted header, otherwise alert created
        if status_code != 202:
            return Response(status=status_code)

        return storage.finish_blob(
            blob=blob,
            digest=digest,
        )

    @never_cache
    def patch(self, request, *args, **kwargs):
        """a patch request is done after a POST with content-length 0 to indicate
        a chunked upload request.
        """
        session_id = kwargs.get("session_id")
        content_length = int(request.META.get("CONTENT_LENGTH"))
        content_range = request.META.get("HTTP_CONTENT_RANGE")
        content_type = request.META.get("CONTENT_TYPE")

        if not session_id or not content_length or not content_type:
            return Response(status=400)

        # If a content range is not defined, assume start to end
        if not content_range:
            content_start = 0
            content_end = content_length - 1

        else:
            # Parse content range into start and end (int)
            try:
                content_start, content_end = parse_content_range(content_range)
            except ValueError:
                return Response(status=400)

        # Confirm that content length (body) == header value, otherwise bad request
        if len(request.body) != content_length:
            return Response(status=400)

        # Get the session id, if it has not expired, keep open for next
        filecache = cache.caches["django_oci_upload"]
        if not filecache.get(session_id):
            return Response(status=400)

        # Break apart into blob id and session uuid
        _, blob_id, version = session_id.split("/")
        try:
            blob = Blob.objects.get(id=blob_id, digest=version)
        except Blob.DoesNotExist:
            return Response(status=404)

        allow_continue, response, _ = is_authenticated(
            request, blob.repository, must_be_owner=True
        )
        if not allow_continue:
            return response

        # Update the blob content_type TODO: There should be some check
        # to ensure that a next chunk content type is not different from that
        # already defined
        blob.content_type = content_type

        # Now process the PATCH request to upload the chunk
        return storage.upload_blob_chunk(
            blob=blob,
            body=request.body,
            content_start=content_start,
            content_end=content_end,
            content_length=content_length,
        )

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="POST",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def post(self, request, *args, **kwargs):
        """POST /v2/<name>/blobs/uploads/"""

        # the name is only used to validate the user has permission to upload
        name = kwargs.get("name")

        # For check media type and content length (if needed)
        content_length = int(request.META.get("CONTENT_LENGTH", 0))
        content_type = request.META.get("CONTENT_TYPE", settings.DEFAULT_CONTENT_TYPE)

        # If no content length, tell the user it's required
        if content_length in [None, ""]:
            return Response(status=411)

        # Validate user having a token, no repository required
        allow_continue, response, user = is_authenticated(
            request, name, repository_exists=False
        )
        if not allow_continue:
            return response

        # Get or create repository
        repository, created = Repository.objects.get_or_create(name=name)

        # If created, add user to owners
        if created:
            repository.owners.add(user)
            repository.save()

        # Otherwise, must be an owner
        else:
            allow_continue, response, _ = is_authenticated(
                request, repository, must_be_owner=True
            )
            if not allow_continue:
                return response

        # Case 1: POST provided with digest == single monolithic upload
        # /v2/<name>/blobs/uploads/?digest=<digest>
        if "digest" in request.GET:

            # Unsupported media type, only needed for digest
            if content_type not in settings.CONTENT_TYPES:
                return Response(status=415)

            digest = request.GET["digest"]

            # Confirm that content length (body) == header value, otherwise bad request
            if len(request.body) != content_length:
                return Response(status=400)

            # The storage.create_blob handles creation of blob with body (no second request required)
            # We only pass the name to return it with the blob's download url, there is no association
            return storage.create_blob(
                body=request.body,
                digest=digest,
                content_type=content_type,
                repository=repository,
            )

        # Case 2; Content type length 0 indicates chunked upload
        elif content_length != 0:
            return storage.create_blob_request(repository)

        # Case 3: POST without digest == single monolithic with POST then PUT
        # returns a session location to upload to with PUT
        return storage.create_blob_request(repository)



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from django.conf.urls import url
from django.urls import path
from django_oci import views
from django_oci import settings

app_name = "django_oci"


urlpatterns = [
    url(
        r"^auth/token/?$",
        views.GetAuthToken.as_view(),
        name="get_auth_token",
    ),
    # https://github.com/opencontainers/distribution-spec/blob/master/spec.md#api-version-check
    url(
        r"^%s/?$" % settings.URL_PREFIX,
        views.APIVersionCheck.as_view(),
        name="api_version_check",
    ),
    url(
        r"^%s/(?P<name>[a-z0-9\/-_]+(?:[._-][a-z0-9]+)*)/tags/list/?$"
        % settings.URL_PREFIX,
        views.ImageTags.as_view(),
        name="image_tags",
    ),
    # This is for a full digest reference
    # https://github.com/opencontainers/distribution-spec/blob/master/spec.md#pulling-an-image-manifest
    url(
        r"^%s/(?P<name>[a-z0-9\/-_]+(?:[._-][a-z0-9]+)*)/manifests/(?P<reference>[A-Za-z0-9_+.-]+:[A-Fa-f0-9]+)/?$"
        % settings.URL_PREFIX,
        views.ImageManifest.as_view(),
        name="image_manifest",
    ),
    # This is for a tag reference
    url(
        r"^%s/(?P<name>[a-z0-9\/-_]+(?:[._-][a-z0-9]+)*)/manifests/(?P<tag>[A-Za-z0-9_+.-]+)/?$"
        % settings.URL_PREFIX,
        views.ImageManifest.as_view(),
        name="image_manifest",
    ),
    url(
        r"^%s/(?P<name>[a-z0-9\/-_]+(?:[._-][a-z0-9]+)*)/blobs/uploads/?$"
        % settings.URL_PREFIX,
        views.BlobUpload.as_view(),
        name="blob_upload",
    ),
    # Listed twice, once with and once without trailing slash
    path(
        "%s/<path:session_id>/blobs/upload/" % settings.URL_PREFIX,
        views.BlobUpload.as_view(),
        name="blob_upload",
    ),
    path(
        "%s/<path:session_id>/blobs/upload" % settings.URL_PREFIX,
        views.BlobUpload.as_view(),
        name="blob_upload",
    ),
    # Also listed twice, once with and once without trailing slash
    path(
        "%s/<path:name>/blobs/<digest>" % settings.URL_PREFIX,
        views.BlobDownload.as_view(),
        name="blob_download",
    ),
    path(
        "%s/<path:name>/blobs/<digest>/" % settings.URL_PREFIX,
        views.BlobDownload.as_view(),
        name="blob_download",
    ),
]



"""

Copyright (c) 2020, Vanessa Sochat

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from django.http.response import Http404, HttpResponse
from django.urls import reverse
from django_oci import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from django_oci.files import ChunkedUpload
from django_oci.models import Blob
from rest_framework.response import Response

import hashlib
import logging
import shutil
import uuid
import os

logger = logging.getLogger(__name__)


def get_storage():
    """Return the correct storage handler based on the key obtained from
    settings
    """
    storage = settings.STORAGE_BACKEND
    lookup = {"filesystem": FileSystemStorage}
    if storage not in lookup:
        logger.warning(
            f"{storage} not supported as a storage backend, defaulting to filesystem."
        )
        storage = "filesystem"
    return lookup[storage]()


# Storage backends for django_oci
# each should have default set of functions for:
# monolithic upload
# chunked upload
# etc.


class StorageBase:
    """A storage base provides shared functions for a storage type"""

    def calculate_digest(self, body):
        """Calculate the sha256 sum for some body (bytes)"""
        hasher = hashlib.sha256()
        hasher.update(body)
        return hasher.hexdigest()


class FileSystemStorage(StorageBase):
    def create_blob_request(self, repository):
        """A create blob request is intended to be done first with a name,
        and content type, and we do all steps of the creation
        except for the actual upload of the file, which is handled
        by a second PUT request to a session id.
        """
        # Generate blob upload session to return to the user
        version = "session-%s" % uuid.uuid4()
        blob = Blob.objects.create(digest=version, repository=repository)
        blob.save()

        # Upon success, the response MUST have a code of 202 Accepted with a location header
        return Response(status=202, headers={"Location": blob.create_upload_session()})

    def finish_blob(
        self,
        blob,
        digest,
    ):
        """Finish a blob, meaning finalizing the digest and returning a download
        url relative to the name provided.
        """

        # In the case of a blob created from upload session, need to rename to be digest
        if blob.datafile.name != digest:
            final_path = os.path.join(
                settings.MEDIA_ROOT, "blobs", blob.repository.name, digest
            )
            if not os.path.exists(final_path):
                dirname = os.path.dirname(final_path)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                shutil.move(blob.datafile.path, final_path)
            else:
                os.remove(blob.datafile.name)
            blob.datafile.name = digest

        # Delete the blob if it already existed
        try:
            existing_blob = Blob.objects.get(repository=blob.repository, digest=digest)
            existing_blob.delete()
        except:
            pass

        blob.digest = digest
        blob.save()

        # Location header must have <blob-location> being a pullable blob URL.
        return Response(status=201, headers={"Location": blob.get_download_url()})

    def create_blob(self, digest, body, content_type, blob=None, repository=None):
        """Create an image blob from a monolithic post. We get the repository
        name along with the body for the blob and the digest.

        Parameters
        ==========
        body (bytes): the request body to write the container
        digest (str): the computed digest of the blob
        content_type (str): the blob content type
        blob (models.Blob): a blob object (if already created)
        """
        # the <digest> MUST match the blob's digest (how to calculate)
        calculated_digest = self.calculate_digest(body)

        # If there is an algorithm prefix, add it
        if ":" in digest:
            calculated_digest = "%s:%s" % (digest.split(":")[0], calculated_digest)

        if calculated_digest != digest:
            return Response(status=400)

        # If we don't have the blob object yet
        created = False
        if not blob:
            blob, created = Blob.objects.get_or_create(
                digest=calculated_digest, repository=repository
            )

        # Delete the blob if it already existed
        try:
            existing_blob = Blob.objects.get(repository=blob.repository, digest=digest)
            existing_blob.delete()
        except:
            pass

        if not blob.datafile:
            datafile = SimpleUploadedFile(
                calculated_digest, body, content_type=content_type
            )
            blob.datafile = datafile

        # The digest is updated here if it was previously a session id
        blob.content_type = content_type
        blob.digest = digest
        blob.save()

        # If it's already existing, return Accepted header, otherwise alert created
        # NOTE: this is set to 201 currently because the conformance test only allows that
        # status_code = 202
        # if created:
        status_code = 201

        # Location header must have <blob-location> being a pullable blob URL.
        return Response(
            status=status_code, headers={"Location": blob.get_download_url()}
        )

    def upload_blob_chunk(
        self,
        blob,
        body,
        content_start,
        content_end,
        content_length,
    ):
        """Upload a chunk of a blob

        Parameters
        ==========
        blob (Blob): the blob to upload to
        body (bytes): the request body to write to the blob
        content_type (str): the blob content type
        content_start (int): the content starting index
        content_end (int): the content ending index
        content_length (int): the content length
        """
        status_code = self.write_chunk(
            blob=blob, content_start=content_start, content_end=content_end, body=body
        )

        # If it's already existing, return Accepted header, otherwise alert created
        if status_code not in [201, 202]:
            return Response(status=status_code)

        # Generate the same upload <location>
        location = reverse(
            "django_oci:blob_upload", kwargs={"session_id": blob.session_id}
        )
        return Response(status=status_code, headers={"Location": location})

    def write_chunk(self, blob, content_start, content_end, body):
        """Write a chunk to a blob. During a chunked upload, the digest corresponds
        to the session_id, and is saved temporarily. It's named on upload finish.
        """
        # Ensure the size is correct (we add 1 to include the start index)
        if len(body) != content_end - content_start + 1:
            return 416

        # If we don't yet have a blob.datafile, create a new one, assert that upload_range starts at 0
        if not blob.datafile:

            # The first request must start at 0
            if content_start != 0:
                return 416

            # Create an empty data file
            datafile = ChunkedUpload(session_id=blob.digest)

        # Uploading another chunk for existing file
        else:
            datafile = ChunkedUpload(session_id=blob.digest, file=blob.datafile.file)

        # Update the chunk, get back the status code
        status_code = datafile.write_chunk(body, content_start)
        blob.datafile.name = datafile.file.name
        blob.save()
        return status_code

    def download_blob(self, name, digest):
        """Given a blob repository name and digest, return response to stream download.
        The repository name is associated to the blob via the image.
        """
        try:
            blob = Blob.objects.get(digest=digest, repository__name=name)
        except Blob.DoesNotExist:
            raise Http404

        if os.path.exists(blob.datafile.name):
            with open(blob.datafile.name, "rb") as fh:
                response = HttpResponse(fh.read(), content_type=blob.content_type)
                response[
                    "Content-Disposition"
                ] = "inline; filename=" + os.path.basename(blob.datafile.name)
                return response

        # If we get here, file doesn't exist
        raise Http404

    def delete_blob(self, name, digest):
        """Given a blob repository name and digest, delete and return success (202)."""
        try:
            blob = Blob.objects.get(digest=digest, repository__name=name)
        except Blob.DoesNotExist:
            raise Http404

        # Delete the blob, will eventually need to check permissions
        blob.delete()
        return Response(status=202)


# Load storage on application init
storage = get_storage()
